from pkg_resources import get_distribution
__version__ = get_distribution('gcloud-aio-auth').version

from gcloud.aio.auth.iam import IamClient
from gcloud.aio.auth.session import AioSession
from gcloud.aio.auth.token import Token
from gcloud.aio.auth.utils import decode
from gcloud.aio.auth.utils import encode
from gcloud.aio.auth.build_constants import BUILD_GCLOUD_REST


__all__ = ['__version__', 'IamClient', 'Token', 'decode', 'encode',
           'AioSession', 'BUILD_GCLOUD_REST']



import datetime

import pytest
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from gcloud.aio.auth import BUILD_GCLOUD_REST
from gcloud.aio.auth import decode
from gcloud.aio.auth import IamClient
from gcloud.aio.auth import Token

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session


@pytest.mark.asyncio
async def test_token_is_created(creds: str) -> None:
    scopes = ['https://www.googleapis.com/auth/taskqueue']

    async with Session() as session:
        token = Token(service_file=creds, session=session, scopes=scopes)
        result = await token.get()

    assert result
    assert token.access_token is not None
    assert token.access_token_duration != 0
    assert token.access_token_acquired_at != datetime.datetime(1970, 1, 1)


@pytest.mark.asyncio
async def test_token_does_not_require_session(creds: str) -> None:
    scopes = ['https://www.googleapis.com/auth/taskqueue']

    token = Token(service_file=creds, scopes=scopes)
    result = await token.get()

    assert result
    assert token.access_token is not None
    assert token.access_token_duration != 0
    assert token.access_token_acquired_at != datetime.datetime(1970, 1, 1)


@pytest.mark.asyncio
async def test_token_does_not_require_creds() -> None:
    scopes = ['https://www.googleapis.com/auth/taskqueue']

    token = Token(scopes=scopes)
    result = await token.get()

    assert result
    assert token.access_token is not None
    assert token.access_token_duration != 0
    assert token.access_token_acquired_at != datetime.datetime(1970, 1, 1)


# Verification code adopted from (adapted to use cryptography lib):
# https://cloud.google.com/appengine/docs/standard/python/appidentity/#asserting_identity_to_third-party_services
async def verify_signature(data, signature, key_name, iam_client):
    key_data = await iam_client.get_public_key(key_name)
    cert = x509.load_pem_x509_certificate(decode(key_data['publicKeyData']),
                                          backend=default_backend())
    pubkey = cert.public_key()

    # raises on failure
    pubkey.verify(decode(signature), data.encode(), padding.PKCS1v15(),
                  hashes.SHA256())


@pytest.mark.asyncio
async def test_sign_blob(creds: str) -> None:
    data = 'Testing Can be confidential!'

    async with Session(timeout=10) as s:
        iam_client = IamClient(service_file=creds, session=s)
        resp = await iam_client.sign_blob(data)
        signed_data = resp['signedBlob']
        await verify_signature(data, signed_data, resp['keyId'], iam_client)


@pytest.mark.asyncio
async def test_get_service_account_public_key_names(creds: str) -> None:
    async with Session(timeout=10) as s:
        iam_client = IamClient(service_file=creds, session=s)
        resp = await iam_client.list_public_keys()
        assert len(resp) >= 1, '0 public keys found.'


@pytest.mark.asyncio
async def test_get_service_account_public_key(creds: str) -> None:
    async with Session(timeout=10) as s:
        iam_client = IamClient(service_file=creds, session=s)
        resp = await iam_client.list_public_keys(session=s)
        pub_key_data = await iam_client.get_public_key(key=resp[0]['name'],
                                                       session=s)

        assert pub_key_data['name'] == resp[0]['name']
        assert 'publicKeyData' in pub_key_data

        key_id = resp[0]['name'].split('/')[-1]
        pub_key_by_key_id_data = await iam_client.get_public_key(key_id=key_id,
                                                                 session=s)

        # Sometimes, one or both keys will be created with "no" expiry.
        pub_key_time = pub_key_data.pop('validBeforeTime')
        pub_key_by_key_id_time = pub_key_by_key_id_data.pop('validBeforeTime')
        assert (pub_key_time == pub_key_by_key_id_time
                or '9999-12-31T23:59:59Z' in {pub_key_time,
                                              pub_key_by_key_id_time})

        assert pub_key_data == pub_key_by_key_id_data



import pytest
from gcloud.aio.auth.build_constants import BUILD_GCLOUD_REST
from gcloud.aio.auth.session import AioSession

if BUILD_GCLOUD_REST:
    import requests

    class Session(requests.Session):
        def __init__(self, *args, **kwargs) -> None:
            self._called = False
            super(requests.Session, self).__init__(*args, **kwargs)

        def close(self) -> None:
            self._called = True
            super(requests.Session, self).close()

        @property
        def closed(self) -> bool:
            return self._called

    requests.Session = Session
else:
    from aiohttp import ClientSession as Session


@pytest.mark.asyncio
async def test_unmanaged_session():
    async with Session() as session:
        gcloud_session = AioSession(session=session)
        assert gcloud_session._shared_session  # pylint: disable=protected-access
        await gcloud_session.close()

        assert not session.closed


@pytest.mark.asyncio
async def test_managed_session():
    gcloud_session = AioSession()
    # create new session
    gcloud_session.session  # pylint: disable=pointless-statement
    if BUILD_GCLOUD_REST:
        gcloud_session._session = Session()  # pylint: disable=protected-access
    assert not gcloud_session._shared_session  # pylint: disable=protected-access
    await gcloud_session.close()

    assert gcloud_session._session.closed  # pylint: disable=protected-access



import json
from typing import Any
from typing import AnyStr
from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import Union

from .build_constants import BUILD_GCLOUD_REST
from .session import AioSession
from .token import Token
from .token import Type
from .utils import encode

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]

API_ROOT_IAM = 'https://iam.googleapis.com/v1'
API_ROOT_IAM_CREDENTIALS = 'https://iamcredentials.googleapis.com/v1'
SCOPES = ['https://www.googleapis.com/auth/iam']


class IamClient:
    def __init__(self, service_file: Optional[Union[str, IO[AnyStr]]] = None,
                 session: Optional[Session] = None,
                 token: Optional[Token] = None) -> None:
        self.session = AioSession(session)
        self.token = token or Token(service_file=service_file,
                                    session=self.session.session,
                                    scopes=SCOPES)

        if self.token.token_type not in {Type.GCE_METADATA,
                                         Type.SERVICE_ACCOUNT}:
            raise TypeError('IAM Credentials Client is only valid for use '
                            'with Service Accounts or GCE Metadata')

    async def headers(self) -> Dict[str, str]:
        token = await self.token.get()
        return {
            'Authorization': f'Bearer {token}',
        }

    @property
    def service_account_email(self) -> Optional[str]:
        return self.token.service_data.get('client_email')

    # https://cloud.google.com/iam/reference/rest/v1/projects.serviceAccounts.keys/get
    async def get_public_key(self, key_id: Optional[str] = None,
                             key: Optional[str] = None,
                             service_account_email: Optional[str] = None,
                             project: Optional[str] = None,
                             session: Optional[Session] = None,
                             timeout: int = 10) -> Dict[str, str]:
        service_account_email = (service_account_email
                                 or self.service_account_email)
        project = project or await self.token.get_project()

        if not key_id and not key:
            raise ValueError('get_public_key must have either key_id or key')

        if not key:
            key = (f'projects/{project}/serviceAccounts/'
                   f'{service_account_email}/keys/{key_id}')

        url = f'{API_ROOT_IAM}/{key}?publicKeyType=TYPE_X509_PEM_FILE'
        headers = await self.headers()

        s = AioSession(session) if session else self.session

        resp = await s.get(url=url, headers=headers, timeout=timeout)

        data: Dict[str, str] = await resp.json()
        return data

    # https://cloud.google.com/iam/reference/rest/v1/projects.serviceAccounts.keys/list
    async def list_public_keys(
            self, service_account_email: Optional[str] = None,
            project: Optional[str] = None,
            session: Optional[Session] = None,
            timeout: int = 10) -> List[Dict[str, str]]:
        service_account_email = (service_account_email
                                 or self.service_account_email)
        project = project or await self.token.get_project()

        url = (f'{API_ROOT_IAM}/projects/{project}/'
               f'serviceAccounts/{service_account_email}/keys')

        headers = await self.headers()

        s = AioSession(session) if session else self.session

        resp = await s.get(url=url, headers=headers, timeout=timeout)

        data: List[Dict[str, Any]] = (await resp.json()).get('keys', [])
        return data

    # https://cloud.google.com/iam/credentials/reference/rest/v1/projects.serviceAccounts/signBlob
    async def sign_blob(self, payload: Optional[Union[str, bytes]],
                        service_account_email: Optional[str] = None,
                        delegates: Optional[List[str]] = None,
                        session: Optional[Session] = None,
                        timeout: int = 10) -> Dict[str, str]:
        service_account_email = (service_account_email
                                 or self.service_account_email)
        if not service_account_email:
            raise TypeError('sign_blob must have a valid '
                            'service_account_email')

        resource_name = f'projects/-/serviceAccounts/{service_account_email}'
        url = f'{API_ROOT_IAM_CREDENTIALS}/{resource_name}:signBlob'

        json_str = json.dumps({
            'delegates': delegates or [resource_name],
            'payload': encode(payload or '').decode('utf-8'),
        })

        headers = await self.headers()
        headers.update({
            'Content-Length': str(len(json_str)),
            'Content-Type': 'application/json',
        })

        s = AioSession(session) if session else self.session

        resp = await s.post(url=url, data=json_str, headers=headers,
                            timeout=timeout)
        data: Dict[str, Any] = await resp.json()
        return data

    async def close(self) -> None:
        await self.session.close()

    async def __aenter__(self) -> 'IamClient':
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()



import pickle

import gcloud.aio.auth.utils as utils  # pylint: disable=unused-import
import pytest


@pytest.mark.parametrize('str_or_bytes', ['Hello Test',
                                          'UTF-8 Bytes'.encode('utf-8'),
                                          pickle.dumps([])])
def test_encode_decode(str_or_bytes):
    encoded = utils.encode(str_or_bytes)
    expected = str_or_bytes
    if isinstance(expected, str):
        try:
            expected = str_or_bytes.encode('utf-8')
        except UnicodeDecodeError:
            pass
    assert expected == utils.decode(encoded)



"""
Google Cloud auth via service account file
"""
import datetime
import enum
import json
import os
import time
from typing import Any
from typing import AnyStr
from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import Union
from urllib.parse import urlencode

import backoff
import cryptography  # pylint: disable=unused-import
import jwt

from .build_constants import BUILD_GCLOUD_REST
from .session import AioSession
# N.B. the cryptography library is required when calling jwt.encrypt() with
# algorithm='RS256'. It does not need to be imported here, but this allows us
# to throw this error at load time rather than lazily during normal operations,
# where plumbing this error through will require several changes to otherwise-
# good error handling.

# Handle differences in exceptions
try:
    # TODO: Type[Exception] should work here, no?
    CustomFileError: Any = FileNotFoundError
except NameError:
    CustomFileError = IOError


# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Response
    from requests import Session
else:
    from aiohttp import ClientResponse as Response  # type: ignore[no-redef]
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]
    import asyncio


GCE_METADATA_BASE = 'http://metadata.google.internal/computeMetadata/v1'
GCE_METADATA_HEADERS = {'metadata-flavor': 'Google'}
GCE_ENDPOINT_PROJECT = (f'{GCE_METADATA_BASE}/project/project-id')
GCE_ENDPOINT_TOKEN = (f'{GCE_METADATA_BASE}/instance/service-accounts'
                      '/default/token?recursive=true')
GCLOUD_TOKEN_DURATION = 3600
REFRESH_HEADERS = {'Content-Type': 'application/x-www-form-urlencoded'}


class Type(enum.Enum):
    AUTHORIZED_USER = 'authorized_user'
    GCE_METADATA = 'gce_metadata'
    SERVICE_ACCOUNT = 'service_account'


def get_service_data(
        service: Optional[Union[str, IO[AnyStr]]]) -> Dict[str, Any]:
    service = service or os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')
    if not service:
        cloudsdk_config = os.environ.get('CLOUDSDK_CONFIG')
        sdkpath = (cloudsdk_config
                   or os.path.join(os.path.expanduser('~'), '.config',
                                   'gcloud'))
        service = os.path.join(sdkpath, 'application_default_credentials.json')
        set_explicitly = bool(cloudsdk_config)
    else:
        set_explicitly = True

    try:
        try:
            with open(service) as f:  # type: ignore[arg-type]
                data: Dict[str, Any] = json.loads(f.read())
                return data
        except TypeError:
            data = json.loads(service.read())  # type: ignore[union-attr]
            return data
    except CustomFileError:
        if set_explicitly:
            # only warn users if they have explicitly set the service_file path
            raise

        return {}
    except Exception:  # pylint: disable=broad-except
        return {}


class Token:
    # pylint: disable=too-many-instance-attributes
    def __init__(self, service_file: Optional[Union[str, IO[AnyStr]]] = None,
                 session: Optional[Session] = None,
                 scopes: Optional[List[str]] = None) -> None:
        self.service_data = get_service_data(service_file)
        if self.service_data:
            self.token_type = Type(self.service_data['type'])
            self.token_uri = self.service_data.get(
                'token_uri', 'https://oauth2.googleapis.com/token')
        else:
            # At this point, all we can do is assume we're running somewhere
            # with default credentials, eg. GCE.
            self.token_type = Type.GCE_METADATA
            self.token_uri = GCE_ENDPOINT_TOKEN

        self.session = AioSession(session)
        self.scopes = ' '.join(scopes or [])
        if self.token_type == Type.SERVICE_ACCOUNT and not self.scopes:
            raise Exception('scopes must be provided when token type is '
                            'service account')

        self.access_token: Optional[str] = None
        self.access_token_duration = 0
        self.access_token_acquired_at = datetime.datetime(1970, 1, 1)

        self.acquiring: Optional[asyncio.Future[Any]] = None  # pylint: disable=unsubscriptable-object,line-too-long

    async def get_project(self) -> Optional[str]:
        project = (os.environ.get('GOOGLE_CLOUD_PROJECT')
                   or os.environ.get('GCLOUD_PROJECT')
                   or os.environ.get('APPLICATION_ID'))

        if self.token_type == Type.GCE_METADATA:
            await self.ensure_token()
            resp = await self.session.get(GCE_ENDPOINT_PROJECT, timeout=10,
                                          headers=GCE_METADATA_HEADERS)

            if not project:
                try:
                    project = await resp.text()
                except (AttributeError, TypeError):
                    project = str(resp.text)

        elif self.token_type == Type.SERVICE_ACCOUNT:
            project = project or self.service_data.get('project_id')

        return project

    async def get(self) -> Optional[str]:
        await self.ensure_token()
        return self.access_token

    async def ensure_token(self) -> None:
        if self.acquiring:
            await self.acquiring
            return

        if not self.access_token:
            self.acquiring = asyncio.ensure_future(self.acquire_access_token())
            await self.acquiring
            return

        now = datetime.datetime.utcnow()
        delta = (now - self.access_token_acquired_at).total_seconds()
        if delta <= self.access_token_duration / 2:
            return

        self.acquiring = asyncio.ensure_future(self.acquire_access_token())
        await self.acquiring

    async def _refresh_authorized_user(self, timeout: int) -> Response:
        payload = urlencode({
            'grant_type': 'refresh_token',
            'client_id': self.service_data['client_id'],
            'client_secret': self.service_data['client_secret'],
            'refresh_token': self.service_data['refresh_token'],
        })

        data: Response = await self.session.post(url=self.token_uri,
                                                 data=payload,
                                                 headers=REFRESH_HEADERS,
                                                 timeout=timeout)
        return data

    async def _refresh_gce_metadata(self, timeout: int) -> Response:
        resp: Response = await self.session.get(url=self.token_uri,
                                                headers=GCE_METADATA_HEADERS,
                                                timeout=timeout)
        return resp

    async def _refresh_service_account(self, timeout: int) -> Response:
        now = int(time.time())
        assertion_payload = {
            'aud': self.token_uri,
            'exp': now + GCLOUD_TOKEN_DURATION,
            'iat': now,
            'iss': self.service_data['client_email'],
            'scope': self.scopes,
        }

        # N.B. algorithm='RS256' requires an extra 240MB in dependencies...
        assertion = jwt.encode(assertion_payload,
                               self.service_data['private_key'],
                               algorithm='RS256')
        payload = urlencode({
            'assertion': assertion,
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
        })

        resp: Response = await self.session.post(self.token_uri, data=payload,
                                                 headers=REFRESH_HEADERS,
                                                 timeout=timeout)
        return resp

    @backoff.on_exception(backoff.expo, Exception, max_tries=5)
    async def acquire_access_token(self, timeout: int = 10) -> None:
        if self.token_type == Type.AUTHORIZED_USER:
            resp = await self._refresh_authorized_user(timeout=timeout)
        elif self.token_type == Type.GCE_METADATA:
            resp = await self._refresh_gce_metadata(timeout=timeout)
        elif self.token_type == Type.SERVICE_ACCOUNT:
            resp = await self._refresh_service_account(timeout=timeout)
        else:
            raise Exception(f'unsupported token type {self.token_type}')

        content = await resp.json()

        self.access_token = str(content['access_token'])
        self.access_token_duration = int(content['expires_in'])
        self.access_token_acquired_at = datetime.datetime.utcnow()
        self.acquiring = None

    async def close(self) -> None:
        await self.session.close()

    async def __aenter__(self) -> 'Token':
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()



import io
import json

import gcloud.aio.auth.token as token
import pytest


@pytest.mark.asyncio
async def test_service_as_io():
    # pylint: disable=line-too-long
    service_data = {
        'type': 'service_account',
        'project_id': 'random-project-123',
        'private_key_id': '399asdfsdf92923k32423a9f9sdf',
        'private_key': '-----BEGIN PRIVATE KEY-----\nABCDF012923949394239492349234923==\n-----END PRIVATE KEY-----\n',
        'client_email': 'gcloud-aio-test@random-project-123.iam.gserviceaccount.com',
        'client_id': '2384283429349234293',
        'auth_uri': 'https://accounts.google.com/o/oauth2/auth',
        'token_uri': 'https://oauth2.googleapis.com/token',
        'auth_provider_x509_cert_url': 'https://www.googleapis.com/oauth2/v1/certs',
        'client_x509_cert_url': 'https://www.googleapis.com/robot/v1/metadata/x509/gcloud-aio%40random-project-123.iam.gserviceaccount.com'
    }

    # io.StringIO does not like str inputs in python2. So in `py3to2` step in CI
    # runs `future-fstrings-show` redefines str literals to unicode, which turns
    # this seemingly noop operation to allow the literal string to get converted
    # to unicode.
    service_file = io.StringIO('{}'.format(json.dumps(service_data)))
    t = token.Token(service_file=service_file,
                    scopes=['https://google.com/random-scope'])

    assert t.token_type == token.Type.SERVICE_ACCOUNT
    assert t.token_uri == 'https://oauth2.googleapis.com/token'
    assert await t.get_project() == 'random-project-123'



import os

import pytest


@pytest.fixture(scope='module')
def creds() -> str:
    return os.environ['GOOGLE_APPLICATION_CREDENTIALS']


@pytest.fixture(scope='module')
def project() -> str:
    return 'dialpad-oss'
